/*
 *	init.c
 *
 *	initialization routines
 *
 *	$Id: init.c,v 1.95 2012/05/04 20:53:10 conrads Exp $
 *
 *****************************************************************************/

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/rtprio.h>
#include <sys/types.h>

#include "mkreadmes.h"

extern	char	categories[];				/* list of categories found in the
												top-level Makefile
											 */
extern	size_t	num_categories;				/* number of categories found */

extern	char	ports_dir[PATH_MAX];		/* top-level ports directory */
extern	char	ports_dir_plus[PATH_MAX];	/* ports_dir + trailing slash */
extern	char	templates_dir[PATH_MAX];	/* directory containing the README.*
												template files
											 */
extern	int		verbose;			/* increase the amount of output */

/* prototypes for global functions */

int		initialize(void);
int		set_ports_dir(const char *path);
int		set_templates_dir(char *path);

/* prototypes for routines in other source files */

extern	char   *resolve_path(const char *path);					/* util.c */
extern	char   *skip_leading(char *str, const char *pattern);	/* util.c */
extern	int		load_index(void);								/* index.c */
extern	int		set_index_categories(void);						/* index.c */
extern	int		validate_path(const char *path);				/* util.c */
extern	void	cleanup(void);									/* util.c */

/*****************************************************************************/

static void set_priority(void)
/*
	Set the program's realtime priority as high as possible

	If error return from rtprio, just ignore and continue running at normal
	priority
*/
{
	struct rtprio rtp;

	rtp.type = RTP_PRIO_REALTIME;
	rtp.prio = 0;

	rtprio(RTP_SET, 0, &rtp);
}

/*****************************************************************************/

static void merge_category(char *category)
/*
 * insert the category (from Makefile.local) into the categories list at its
 * properly sorted location
 */
{
	char   *categories_ptr = categories;
	char   *ptr;
	int		n;		/* number of bytes we have to shift upwards for an insert */

	while (*categories_ptr)
	{
		if (strcmp(category, categories_ptr) < 0)	/* time to insert! */
		{
			/* move everything up to make room for new category */

			/* first, figure out how many bytes we need to move */
			ptr = categories_ptr;

			while (*ptr)
			{
				ptr += strlen(ptr) + 1;
			}

			n = ptr - categories_ptr;	/* number of bytes we have to move */

			ptr = categories_ptr + strlen(category) + 1;	/* base of move destination */

			memmove(ptr, categories_ptr, n);	/* move it! */

			strcpy(categories_ptr, category);	/* copy in the new category */
			return;
		}

		categories_ptr += strlen(categories_ptr) + 1;
	}

	/* we reached the end of the list, just append the new category */
	/* copy category to the categories list and bump pointer to next
		free slot
	 */
	categories_ptr = stpcpy(categories_ptr, category) + 1;
	*categories_ptr = '\0';		/* add the extra nul end-of-list marker */
}

/*****************************************************************************/

static int gen_categories_list(void)
/*
 *	generate a list of all ports categories from the top-level Makefile
 *	and Makefile.local (if it exists)
 *
 *	returns number of categories found, or -1 on error
 */
{
	char	Makefile_path[PATH_MAX];
	char	Makefile_buf[LINE_MAX];
	FILE   *Makefile;
	char   *category;		/* pointer to category in current input buffer */

	char   *categories_ptr = categories;	/* pointer into categories list */

	/* open the top-level Makefile for reading */
	strcpy(Makefile_path, ports_dir);
	strcat(Makefile_path, MAKEFILENAME);

	if ((Makefile = fopen(Makefile_path, "r")) == NULL)
	{
		perror(Makefile_path);
		fprintf(stderr, "Unable to open top-level ports Makefile\n");
		return -1;
	}

	/* scan through all the "SUBDIR += " lines to get all the categories */
	while (fgets(Makefile_buf, LINE_MAX, Makefile) != NULL)
	{
		if ((category = skip_leading(Makefile_buf, "SUBDIR += ")) !=
				Makefile_buf)	/* found a SUBDIR line */
		{
			/* don't want the newline */
			category[strlen(category) - 1] = '\0';

			/* copy category to the categories list and bump pointer to next
				free slot
			 */
			categories_ptr = stpcpy(categories_ptr, category) + 1;
			++num_categories;
		}
	}
	/* add an extra nul character to mark the end of the categories list */
	*categories_ptr = '\0';

	fclose(Makefile);

	/* try to open the top-level Makefile.local for reading
	 *
	 * Makefile.local is optional, not mandatory, so don't complain
	 * register an error if unable to open it
	 */
	strcpy(Makefile_path, ports_dir);
	strcat(Makefile_path, MAKEFILELOCALNAME);

	if ((Makefile = fopen(Makefile_path, "r")) != NULL)
	{
		/* scan through all the "SUBDIR += " lines to get all the categories */
		while (fgets(Makefile_buf, LINE_MAX, Makefile) != NULL)
		{
			if ((category = skip_leading(Makefile_buf, "SUBDIR += ")) !=
					Makefile_buf)	/* found a SUBDIR line */
			{
				/* don't want the newline */
				category[strlen(category) - 1] = '\0';

				/* merge category into the categories list at its properly
					sorted location
				 */
				merge_category(category);
				++num_categories;
			}
		}

		fclose(Makefile);
	}
	return 0;
}

/*****************************************************************************/

int set_ports_dir(const char *path)
/*
	set the ports_dir variable, a character array containing the name of the
	top-level ports directory

	set ports_dir to 'path' if not NULL, after running some validity checks
	on it first

	if 'path' is NULL, set ports_dir to the current setting of the PORTSDIR
	environment variable, if defined (again, after running some validity
	checks on it), or the default of PORTSTOP ("/usr/ports")

	set ports_dir_plus to ports_dir plus a trailing slash

	returns 0 on success, -1 on error
 */
{
	char   *portsdir_env;	/* current setting of PORTSDIR */
	char   *ptr;			/* temporary general-purpose pointer */

	/*
	 *	First case: (path != NULL), meaning user selected the "-p pathname"
	 *	command line option, which will override any previous setting of the
	 *	ports_dir variable, as well as the PORTSDIR environment variable
	 */
	if (path != NULL)
	{
		if (strlen(ports_dir))
		{
			/* advise user of the change of path */
			fprintf(stderr,
					"Changing top-level directory from \"%s\" to \"%s\"\n",
					ports_dir, path);
		}

		if ((ptr = resolve_path(path)) == NULL)
		{
			fprintf(stderr, "Invalid pathname argument: \"%s\"\n", path);
			return -1;
		}

		if (validate_path(ptr) != 0)
		{
			free(ptr);
			return -1;
		}
		else
		{
			strcpy(ports_dir, ptr);
			free(ptr);
		}
	}
	else
	/*
	 *	The other case: (path == NULL), meaning we were called from
	 *	initialize() with a NULL path
	 */
	if ((portsdir_env = getenv("PORTSDIR")) != NULL)
	{
		if ((ptr = resolve_path(portsdir_env)) == NULL)
		{
			fprintf(stderr,
					"Invalid pathname in PORTSDIR environment variable: \"%s\"\n",
					portsdir_env);
			return -1;
		}

		if (validate_path(ptr) != 0)
		{
			fprintf(stderr, "Unusable PORTSDIR setting\n");
			free(ptr);
			return -1;
		}
		else
		{
			strcpy(ports_dir, ptr);
			free(ptr);
			if (verbose)
			{
				fprintf(stderr,
						"Using PORTSDIR (\"%s\") as top-level directory\n",
						ports_dir);
			}
		}
	}
	else
	{
		strcpy(ports_dir, PORTSTOP);	/* use default top */
		if (verbose)
			fprintf(stderr, "Using default top-level directory \""
					PORTSTOP "\"\n");
	}

	if (verbose && (path != NULL))
		fprintf(stderr, "Setting top-level directory to \"%s\"\n",
				ports_dir);

	/* now set ports_dir_plus to whatever we got for ports_dir, and add a
		trailing slash
	 */
	strcpy(ports_dir_plus, ports_dir);
	strcat(ports_dir_plus, "/");

	return 0;
}

/*****************************************************************************/

int set_templates_dir(char *path)
/*
	set the templates_dir variable

	returns 0 on success, -1 on error
 */
{
	char *ptr;

	if (path != NULL)	/* called from main() by the "-t" option */
	{
		if ((ptr = resolve_path(path)) == NULL)
		{
			fprintf(stderr, "Invalid pathname argument: \"%s\"\n", path);
			return -1;
		}

		if (validate_path(ptr) != 0)
		{
			free(ptr);
			return -1;
		}
		/* else */
		strcpy(templates_dir, ptr);
		free(ptr);
	}
	else	/* called from initialize() with NULL path argument */
	{
		/* construct the path to the default Templates directory,
			respecting the PREFIX where the program was originally
			installed
		 */
		strcpy(templates_dir, PREFIX);
		strcat(templates_dir, TEMPLATESDIR);

		if (validate_path(templates_dir) != 0)	/* something's amiss(ing) */
		{
			return -1;
		}

		if (verbose)
			fprintf(stderr, "Using default templates directory \"%s\"\n",
					templates_dir);
	}

	if (verbose && (path != NULL))
		fprintf(stderr, "Setting templates directory to \"%s\"\n",
				templates_dir);

	return 0;
}

/*****************************************************************************/

int initialize(void)
/*
	Initialize the program by performing the following series of operations:

	Attempt to elevate the program's realtime priority as high as possible.
	If setting the priority fails, just ignore it and run at normal priority.

	Register a cleanup routine to be called automatically when the program
	exits, to ensure that we free up whatever memory we obtain via malloc()

	Unless it was already set with the "-p pathname" command line option,
	set the ports_dir variable to the top-level ports directory, either
	from the environment variable PORTSDIR (if set), or the #defined value of
	PORTSTOP ("/usr/ports/").

	Unless it was already set with the "-t pathname" command line option,
	set the templates_dir variable to the default templates directory
	(${PREFIX}/share/mkreadmes/Templates).

	Generate a list of categories from the top-level Makefile.

	Load the index file into a dynamically allocated buffer.  The
	load_index_file() routine will also generate a list of pointers to the
	start of each category's section within the index buffer, to optimize
	ports searches later

	Load each of the three template files into their own dynamically allocated
	buffers.

	Returns 0 on success, or -1 on error
*/
{
	/* set program's realtime priority as high as possible */

	if (getuid() == 0)	/* only try if running as root */
	{
		set_priority();
	}

	/* register a cleanup routine with atexit(3) to automatically
		free whatever memory we allocated when the program exits
	 */
	if (atexit(cleanup) == -1)
	{
		perror("atexit() register of cleanup routine failed");
		return -1;
	}

	/* set the global variables ports_dir and ports_dir_plus (ports_dir with a
		trailing slash), if not already set by the "-p" option
	 */
	if (strlen(ports_dir) == 0)
	{
		if (set_ports_dir(NULL) == -1)
		{
			return -1;
		}
	}

	/* set the global variable templates_dir,
		if not already set by the "-t" option
	 */
	if (strlen(templates_dir) == 0)
	{
		if (set_templates_dir(NULL) == -1)
		{
			return -1;
		}
	}
	/* generate a list of categories from the top-level Makefile */
	if (gen_categories_list() != 0)
		return -1;
	/* load the index file into memory */
	if (load_index() == -1)
	{
		fprintf(stderr, "Unable to load the index file!\n");
		return -1;
	}
	/* setup an array of pointers to the first index line of each category */
	return set_index_categories();
}

