/*
 *	File: mkreadmes.c
 *
 *	Contains three routines that drive the body of the work, one for
 *	each level of readmes -- the top level, the category level, and
 *	the individual port level -- plus a few small auxiliary functions
 *
 *	main() will call only one of the make_readme_* routines, based on whether
 *	the pathname given as a command line argument (or the top-level ports
 *	directory, if none is provided) is determined to be at the top, category,
 *	or individual port level
 *
 *	$Id: mkreadmes.c,v 1.128 2012/05/04 20:53:10 conrads Exp $
 *
 *****************************************************************************/

#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#include "mkreadmes.h"

/* placeholder names used in templates.  on a successful match, the
	placeholder's index in this array is then used as an index into the
	index_field[] array to obtain a pointer to the corresponding index field's
	data, which will have been safely tucked away in the index_line[] array
	(an array of nul-terminated strings) by the split_index_line() routine
 */
static const char *placeholder[] = {
	"%%PKG%%",
	"%%PORT%%",
	"%%PREFIX%%",
	"%%COMMENT%%",
	"%%DESCR%%",
	"%%EMAIL%%",
	"%%CATEGORY%%",
	"%%BUILD_DEPENDS%%",
	"%%RUN_DEPENDS%%",
	"%%WEBSITE%%",
	"%%TOP%%",
	"%%SUBDIR%%",
	NULL			/* end of list marker */
};

static	char	category_comment[LINE_MAX];	/* comment from a category's
												Makefile, saved by
												make_readme_category() for use
												by make_readme_top()
											 */
char	categories[MAXCATLIST];	/* list of categories found in the top-level
									Makefile
								 */
size_t	num_categories = 0;		/* number of categories found */

char *(*index_categories)[] = NULL;	/* pointer to an array of pointers to
										the first line of each category's
										section within the index buffer
									 */

extern	char	*index_field[];	/* pointers to current port's index fields */

char   *next_index;			/* pointer to the next index line to search */
char   *next_cat;			/* pointer to first index line of next category */

char	templates_dir[PATH_MAX];	/* directory containing the README.*
										template files
									 */
/* pointers to dynamically allocated buffers for the template files */
char	*template_buf_top = NULL;
char	*template_buf_category = NULL;
char	*template_buf_port = NULL;

extern	char	ports_dir[PATH_MAX];		/* top-level ports directory */
extern	char	ports_dir_plus[PATH_MAX];	/* ports_dir plus trailing slash */

extern	int		quiet;			/* silence normal output */
extern	int		verbose;		/* increase the amount of output */

/* prototypes for global functions */

int make_readme(const char *path, int make_level);

/* prototypes for external functions */

extern	char   *load_file(const char *path);						/* util.c */
extern	char   *skip_leading(const char *str, const char *pattern);	/* util.c */
extern	int		search_index(const char *path);					   /* index.c */

/*****************************************************************************/

static char *load_template(const char *ext)
/*
	load the template file with extension 'ext' into a dynamically allocated
	buffer

	returns a pointer to the buffer on success, or NULL on failure

	Note: This function makes no attempt to determine whether the requested
	file has already been loaded, relying instead on the caller to have
	already checked for that before calling this function.  Both it and its
	auxiliary routine, load_file() in util.c, don't know and don't care whether
	the pointer variable the result will be assigned to on return has already
	been set, so use with caution!

	There *is* an upside to this methodology: because of the lack of checking
	before proceeding to load the file, both of these routines are very simple,
	and very fast.  :-)
 */
{
	char	template_path[PATH_MAX];

	/* generate the fully qualified pathname for the template file */
	strcpy(template_path, templates_dir);
	strcat(template_path, TEMPLATENAME);	/* template name minus extension */

	/* add extension to pathname */
	strcat(template_path, ext);

	return load_file(template_path);
}

/*****************************************************************************/

static int load_templates(void)
/*
	load all three of the template files -- README.port, README.category and
	README.top -- into memory

	this is just a "convenience" function, called from make_readme
	for top-level builds

	returns 0 on success, or -1
*/
{
	/* Don't load any files that have already been loaded */

	if (template_buf_port == NULL)		/* not loaded yet */
		if ((template_buf_port = load_template("port")) == NULL)
			return -1;

	if (template_buf_category == NULL)	/* same here */
		if ((template_buf_category = load_template("category")) == NULL)
			return -1;

	if (template_buf_top == NULL)		/* and here */
		if ((template_buf_top = load_template("top")) == NULL)
			return -1;

	return 0;
}

/*****************************************************************************/

static inline char *HTMLify(char *dest, char *src )
/*
	Copy 'src' to 'dest', replacing any occurences of '&', '<' or '>'
	with their HTML representations

	Returns a pointer to the location in 'dest' one past the last
	character copied

	Uses the addstr() macro defined in mkreadmes.h
 */
{
	while (*src)
	{
		switch(*src)
		{
			case '&':
			{
				addstr(dest, "&amp;");
				++src;
				continue;
			}
			case '<':
			{
				addstr(dest, "&lt;");
				++src;
				continue;
			}
			case '>':
			{
				addstr(dest, "&gt;");
				++src;
				continue;
			}
			default:
				*dest++ = *src++;
		}
	}

	return dest;
}

/*****************************************************************************/

static inline int match_placeholder(const char *ptr)
/*
	Try to find a matching "%%VARIABLE%%" within the placeholder array
	for the string pointed to by 'ptr'

	On entry to this routine, 'ptr' will be pointing to the first "%" in the
	possible symbol name we're looking for

	Returns a positive integer representing the variable's index in the
	placeholder array, or -1 if no match is found
 */
{
	int j;

	int i = 0;

	while (placeholder[i])
	{
		j = 0;

		while(ptr[j] == (placeholder[i])[j])
			++j;

		if ((placeholder[i])[j] == '\0')
			return i;
		++i;
	}

	/* No match found */
	return -1;
}

/*****************************************************************************/

static inline int category_index(const char *cat)
/*
	scan the list of category names to get an index to use with the
	index_categories[] array, from which we'll obtain a pointer to
	the first line of this category in the index buffer

	returns the index at which the category was found in the list, or
	num_categories if not found
*/
{
	char   *categories_ptr = categories;
	int		i = 0;

	while (*categories_ptr)
	{
		if (strcmp(cat, categories_ptr) == 0)
			return i;									/* found a match */
		/* else */
		categories_ptr += strlen(categories_ptr) + 1;	/* bump pointer and */
		++i;											/* index */
	}

	return i;	/* (i == num_categories) = not found */
}

/*****************************************************************************/

static int make_readme_port(const char *path, int make_level)
/*
	Make the README.html file for a single port

	On entry, category will already be pointing either to the "category/port"
	part of the port's pathname (if make_level == PORT) or just the category
	part (if make_level == CATEGORY or TOP)

	Returns 0 on success, or 1 on failure

	Uses the addstr() macro defined in mkreadmes.h
*/
{
	char	readme_buf[MAXREADME_PORT];	/* output buffer */
	char	readme_path[PATH_MAX];		/* output pathname */
	int		readme;						/* file descriptor */
	char   *catport;					/* category/port portion of path */
	char	catonly[MAXCATNAME];		/* just the category name */
	char   *category;					/* pointer to category name */
	size_t	catlen;						/* length of category name */
	int		i;					/* index into the placeholder array where a
									template match was found
								 */
	int		cat_index;	/* index into the categories[] and index_categories[]
							arrays for this port's category (can't initialize
							here as we do in make_readme_category() below, not
							until we make certain that 'category' has been set
							properly
						 */

	char   *readme_ptr = readme_buf;	/* pointer into output buffer */
	char   *template_ptr = template_buf_port;	/* pointer into template
													buffer
												 */

	/* save a pointer to the category/port portion of port's path */
	catport = skip_leading(path, ports_dir_plus);

	/* copy just the category to catonly */
	catlen = strcspn(catport, "/");
	strncpy(catonly, catport, catlen);
	catonly[catlen] = '\0';
	category = catonly;			/* set category to point to the string we
									just copied
								 */
	if (make_level == PORT)	/* need to set next_index and next_cat */
	{
	/* set next_index to point to the first line of this port's category's
		section in the index buffer, so we won't waste any time searching in
		any of the preceding categories
	 */
		next_index = (*index_categories)[cat_index = category_index(category)];

		if (next_index == NULL)	/* category not found :-( */
		{
			fprintf(stderr, "No such category: %s\n", category);
			return 1;
		}

		/* set next_cat to point to the first index line of the next category,
			to use as the search cutoff point
		 */
		next_cat = (*index_categories)[cat_index + 1];
	}

	/* Find the port's entry in the index file, saving pointers to its index
		information in the index_field[] array
	 */
	if (!search_index(path))
	{
		fprintf(stderr, "Port %s not found in INDEX\n", catport);
		return 1;
	}

	/* For efficiency's sake, don't output the following message
		unless this one port is all we're building, or the verbose
		flag is set to >= 3
	 */
	if ((verbose > 2) || ((make_level == PORT) && !quiet))
		printf("Making README.html for port %s\n", catport);

	/* Copy the template buffer to the readme buffer,
		performing any substitutions needed
	 */

	while (*template_ptr)
	{
		/* just copy from template buffer to readme buffer
			unless we hit a '%'
		 */
		if ((*template_ptr == '%') &&
			((i = match_placeholder(template_ptr)) >= 0))
		{
			/* a few fields require special handling:

				for the PORT field, use the "category/port" string we saved
				earlier in the variable catport[]

				for the COMMENT field, convert any '&', '<' or '>' characters
				into their HTML representations

				for the DESCR field, adjust the path prefix to the realpath
				of the top of our ports tree, which may differ from the
				canonical "/usr/ports" used in the index field

				for BUILD_DEPENDS and RUN_DEPENDS, don't output anything
				if these fields are empty in the INDEX file, otherwise,
				precede them with the introductory text #defined in
				BUILD_DEPS or RUN_DEPS

				for WEBSITE, output the lead-in text #defined in WWW_BEFORE,
				then the field contents, and follow it up with the text
				#defined in WWW_AFTER

			 */
			switch (i)
			{
				case PKG:
					addstr(readme_ptr, index_field[i]);
					break;
				case PORT:
					addstr(readme_ptr, catport);
					break;
				case COMMENT:
					/* the port's comment may have some bad HTML chars */
					readme_ptr = HTMLify(readme_ptr, index_field[COMMENT]);
					break;
				case DESCR:
					addstr(readme_ptr, ports_dir);
					addstr(readme_ptr, skip_leading(index_field[i], PORTSTOP));
					break;
				case EMAIL:
					addstr(readme_ptr, index_field[i]);
					break;
				case BUILD_DEPENDS:
					if (strlen(index_field[i]))
					{
						addstr(readme_ptr, BUILD_DEPS);
						addstr(readme_ptr, index_field[i]);
					}
					break;
				case RUN_DEPENDS:
					if (strlen(index_field[i]))
					{
						addstr(readme_ptr, RUN_DEPS);
						addstr(readme_ptr, index_field[i]);
					}
					break;
				case WEBSITE:
					if (strlen(index_field[i]))
					{
						addstr(readme_ptr, WWW_BEFORE);
						addstr(readme_ptr, index_field[i]);
						addstr(readme_ptr, WWW_AFTER);
					}
					break;
				case TOP:
					addstr(readme_ptr, ports_dir);
					break;
			}
			/* bump template pointer past placeholder */
			template_ptr += strlen(placeholder[i]);
		}
		else *readme_ptr++ = *template_ptr++;
	}

	/* Open the port's README.html file for writing */

	strcpy(readme_path, path);			/* port's path in the ports tree */
	strcat(readme_path, READMENAME);	/* append "/README.html" */

	if ((readme = open(readme_path, O_WRONLY|O_CREAT|O_TRUNC|O_DIRECT, MODE))
				== -1)
	{
		perror(readme_path);
		return 1;
	}

	/* write readme_buf to output file and close */

	if (write(readme, readme_buf, (size_t)(readme_ptr - readme_buf)) == -1)
	{
		perror(readme_path);
		close(readme);
		return 1;
	}

	close(readme);

	return 0;
}

/*****************************************************************************/

static int make_readme_category(const char *path, int make_level)
/*
	Make the README.html file for a single category

	Calls make_readme_port() to build the README.html file for each
	port within the category

	Returns 0 on success, or 1 on failure

	Uses the addstr() macro defined in mkreadmes.h
*/
{
	char	readme_buf[MAXREADME_CAT];	/* output buffer */
	char	readme_path[PATH_MAX];		/* output pathname */
	int		readme;						/* file descriptor */

	char	Makefile_buf[LINE_MAX];		/* input buffer */
	char	Makefile_path[PATH_MAX];	/* input pathname */
	FILE   *Makefile;					/* input stream for category's
											Makefile
										 */
	char	ports[MAXPORTSLIST];		/* list of ports found in a category's
											Makefile
										 */
	char   *port;				/* pointer to port in current input buffer */
	char	port_path[PATH_MAX];	/* full path to port in ports tree */

	int		i;					/* index into the placeholder array where a
									template match was found
								 */

	char   *comment_ptr = Makefile_buf;	/* for locating the COMMENT line */
	char   *readme_ptr = readme_buf;	/* pointer into output buffer */
	char   *ports_ptr = ports;			/* pointer into ports list */
	char   *template_ptr = template_buf_category;	/* pointer into template
														buffer
													 */
	int		num_ports = 0;				/* number of ports in category */

/* get the index into the categories[] and index_categories[] arrays
	for this category
 */
	char   *category = skip_leading(path, ports_dir_plus);
	int		cat_index = category_index(category);

	/* set next_index to point to the first line of this category's section
		in the index buffer, so we won't waste any time searching in any of
		the preceding categories
	 */
	if ((next_index = (*index_categories)[cat_index]) == NULL)	/* not found */
	{
		fprintf(stderr, "No such category: %s\n", category);
		return 1;
	}

	/* set next_cat to point to the first index line of the next category,
		to use as the search cutoff point
	 */
	next_cat = (*index_categories)[cat_index + 1];

	/* Open the category's Makefile for reading */

	strcpy(Makefile_path, path);
	strcat(Makefile_path, MAKEFILENAME);

	if ((Makefile = fopen(Makefile_path, "r")) == NULL)
	{
		perror(Makefile_path);
		return 1;
	}

	/* First, find the COMMENT line, saving it externally in category_comment[]
		in case we were called from make_readme_top(), which will need the
		comment for its own README.html file
	 */

	while ((fgets(Makefile_buf, LINE_MAX, Makefile) != NULL))
	{
		if ((comment_ptr = skip_leading(Makefile_buf, "    COMMENT = "))
			!= Makefile_buf)	/* found it */
		{
			/* don't want the newline */
			comment_ptr[strlen(comment_ptr) - 1] = '\0';
			strcpy(category_comment, comment_ptr);	/* save the COMMENT */
			break;
		}
	}

	/* Got the comment, continue reading to get all the SUBDIRs */

	while ((fgets(Makefile_buf, LINE_MAX, Makefile) != NULL))
	{
		if ((port = skip_leading(Makefile_buf, "    SUBDIR += ")) !=
				Makefile_buf)	/* found a SUBDIR line */
		{
			/* don't want the newline */
			port[strlen(port) - 1] = '\0';
			ports_ptr = stpcpy(ports_ptr, port) + 1;	/* add to ports list */
			++num_ports;
		}
	}

	/* add an extra nul character to mark the end of the ports list */
	*ports_ptr = '\0';

	fclose(Makefile);

	/* Done collecting the data, now we can start generating
		the README.html file
	 */
	if (!quiet)
		printf("Making README.html for category %s (%d ports)\n",
			category, num_ports);

	/* Copy the template buffer to the readme buffer,
		performing any substitutions needed
	 */

	while (*template_ptr)
	{
		/* just copy from template buffer to readme buffer
			unless we hit a '%'
		 */
		if ((*template_ptr == '%') &&
			((i = match_placeholder(template_ptr)) >= 0))
		{
			/* SUBDIR requires special handling, calling make_readme_port()
				for each port in the category to build the port's README.html
				file and to obtain its package name and comment for the
				category README.html file

				just as in make_readme_port(), for the port's COMMENT field,
				convert any '&', '<' or '>' characters into their HTML
				representations
			 */
			switch (i)
			{
				case COMMENT:	/* the category's comment, not a port's */
					addstr(readme_ptr, category_comment);
					/* add the number of ports in category to end of comment */
					readme_ptr += sprintf(readme_ptr, " (%d ports)", num_ports);
					break;
				case CATEGORY:
					addstr(readme_ptr, category);
					break;
				case SUBDIR:
					ports_ptr = ports;

					while (*ports_ptr)
					{
						port = ports_ptr;	/* port's base name */

						/* generate the full path to the port */
						strcpy(port_path, path);	/* category's path */
						strcat(port_path, "/");
						strcat(port_path, port);	/* append port's name */

						/* make the port's README.html */
						make_readme_port(port_path, make_level);

						/* now we can use the information left over in the
							index_field[] array by make_readme_port() */
						addstr(readme_ptr, "<a href=\"");
						addstr(readme_ptr, port);
						addstr(readme_ptr, "/README.html\">");
						addstr(readme_ptr, index_field[PKG]);
						addstr(readme_ptr, "</a>: ");

						/* the port's comment may have some bad HTML chars */
						readme_ptr = HTMLify(readme_ptr, index_field[COMMENT]);
						strcpy(readme_ptr++, "\n");

						/* bump ports_ptr to next item in ports array */
						ports_ptr += strlen(port) + 1;
					}
			}
			/* bump template pointer past placeholder */
			template_ptr += strlen(placeholder[i]);
		}
		else *readme_ptr++ = *template_ptr++;
	}

	/* Open the category's README.html file for writing */

	strcpy(readme_path, path);
	strcat(readme_path, READMENAME);

	if ((readme = open(readme_path, O_WRONLY|O_CREAT|O_TRUNC|O_DIRECT, MODE))
				== -1)
	{
		perror(readme_path);
		return 1;
	}

	/* write readme_buf to output file and close */

	if (write(readme, readme_buf, (size_t)(readme_ptr - readme_buf)) == -1)
	{
		perror(readme_path);
		close(readme);
		return 1;
	}

	close(readme);

	return 0;
}

/*****************************************************************************/

static int make_readme_top(const char *path, int make_level)
/*
	Make the top-level README.html file

	Calls make_readme_category() to build the README.html file for each
	category in the categories list created earlier from the top-level Makefile

	Returns 0 on success, or 1 on failure

	Uses the addstr() macro defined in mkreadmes.h
*/
{
	char	readme_buf[MAXREADME_TOP];	/* output buffer */
	char	readme_path[PATH_MAX];		/* output pathname */
	int		readme;						/* file descriptor */

	char   *category;				/* name of current category */
	char	category_path[PATH_MAX];	/* fully qualified path to current
											category
										 */
	int		i;						/* index into the placeholder array
										where a template match was found
									 */
	char   *categories_ptr = categories;	/* pointer into categories list */
	char   *readme_ptr = readme_buf;	/* pointer into output buffer */
	char   *template_ptr = template_buf_top;	/* pointer into template buffer
												 */
	if (!quiet)
		printf("Making top-level README.html in %s (%lu categories)\n",
			ports_dir, num_categories);

	/* Copy the template buffer to the readme buffer,
		performing any substitutions needed
	 */

	while (*template_ptr)
	{
		/* just copy from template buffer to readme buffer
			unless we hit a '%'
		 */
		if ((*template_ptr == '%') &&
			((i = match_placeholder(template_ptr)) >= 0))
		{
			/* %%SUBDIR%% is the only placeholder used in the top-level
				template, and will cause this routine to call
				make_readme_category() for each category found in the top-level
				Makefile (already obtained and saved earlier in the categories[]
				list), which in turn will call make_readme_port() for each port
				in the category's own Makefile.

				On return from make_readme_category() the category's COMMENT
				will be stored in category_comment[] for use in the top-level
				README.html file.
			 */
			if (i == SUBDIR)
			{
				categories_ptr = categories;

				while (*categories_ptr)
				{
					category = categories_ptr;	/* base name of category */

					/* generate the full path to the category */
					strcpy(category_path, ports_dir_plus);
					strcat(category_path, category);

					/* make the category's README.html, saving its
						COMMENT in category_comment[] */
					make_readme_category(category_path, make_level);

					/* now we have all the information we need
						for this category
					 */
					addstr(readme_ptr, "<a href=\"");
					addstr(readme_ptr, category);
					addstr(readme_ptr, "/README.html\">");
					addstr(readme_ptr, category);
					addstr(readme_ptr, "</a>: ");
					addstr(readme_ptr, category_comment);
					strcpy(readme_ptr++, "\n");

					/* bump pointer to next item in categories array */
					categories_ptr += strlen(category) + 1;
				}
			}
			/* bump template pointer past placeholder */
			template_ptr += strlen(placeholder[i]);
		}
		else *readme_ptr++ = *template_ptr++;
	}

	/* Open the top-level README.html file for writing */

	strcpy(readme_path, path);			/* fully qualified top-level path */
	strcat(readme_path, READMENAME);	/* append "/README.html" */

	if ((readme = open(readme_path, O_WRONLY|O_CREAT|O_TRUNC|O_DIRECT, MODE))
				== -1)
	{
		perror(readme_path);
		return 1;
	}

	/* write readme_buf to output file and close */

	if (write(readme, readme_buf, (size_t)(readme_ptr - readme_buf)) == -1)
	{
		perror(readme_path);
		close(readme);
		return 1;
	}

	close(readme);

	return 0;
}

/*****************************************************************************/

int make_readme(const char *path, int make_level)
/*
	wrapper for the actual make_readme_* routines, called from main()

	make sure the template files are loaded for whichever level of build
	we're doing, then do the actual build

	returns the return code of whichever make_readme_* routine it calls

	(the default case in the switch statement below should never happen,
	and is provided merely to avoid a compiler warning)
 */
{
	switch (make_level)
	{
		case TOP:
		{
			/* load_templates() will check first before trying to load a
				a template file, to see if it has already been loaded, and
				if so, won't try to load it again
			 */
			if (load_templates() != 0)
				return -1;
			/* else */
			return make_readme_top(path, make_level);
		}

		/* for CATEGORY and PORT, we do our checking right here first, before
			calling load_template()
		 */

		case CATEGORY:
		{
			if (template_buf_category == NULL)	/* not loaded yet */
				if ((template_buf_category = load_template("category")) == NULL)
					return -1;
			if (template_buf_port == NULL)		/* ditto */
				if ((template_buf_port = load_template("port")) == NULL)
					return -1;
			return make_readme_category(path, make_level);
		}

		case PORT:
		{
			if (template_buf_port == NULL)	/* and likewise for PORT */
				if ((template_buf_port = load_template("port")) == NULL)
					return -1;
			return make_readme_port(path, make_level);
		}
		default:	return -1;	/* this should never happen, but let's
									keep the compiler happy
								 */
	}
}

